{--
    Copyright © 2011, Ingo Wechsung
 *
    All rights reserved.
 *
    Redistribution and use in source and binary forms, with or
    without modification, are permitted provided that the following
    conditions are met:

    -   Redistributions of source code must retain the above copyright
        notice, this list of conditions and the following disclaimer.

    -   Redistributions in binary form must reproduce the above
        copyright notice, this list of conditions and the following
        disclaimer in the documentation and/or other materials provided
        with the distribution. Neither the name of the copyright holder
        nor the names of its contributors may be used to endorse or
        promote products derived from this software without specific
        prior written permission.

    *THIS SOFTWARE IS PROVIDED BY THE
    COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY EXPRESS OR
    IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
    WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
    PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER
    OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
    SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
    LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF
    USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED
    AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
    LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING
    IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF
    THE POSSIBILITY OF SUCH DAMAGE.*
-}

{--
 * Support for ad hoc parallelism in Frege through java's ForkJoin mechanism
 -}

package frege.lib.ForkJoin where

infixr 2 `par`   -- like seq
infixr 1 `$|`    -- like ($)
infixr 4 `:|`    -- like (:)

{--
    Native operation to fork evaluation of some lazy value.

    A ForkJoinTask will be created and forked unless
    the value is already evaluated.

    When the ForkJoinTask begins execution it will invoke the @call()@ method
    of the @frege.runtime.Lazy@ instance that was passed as value.
    The sole purpose of @call@ is to make an @Lazy@ implement the
    Java @Callable@ interface, it just evaluates the lazy value in the
    usual way. Because evaluation is synchronized, any attempts to evaluate the
    value from other tasks will block until evaluation is complete and the
    @Lazy@ is updated.

    We pass the value as partial applied function to avoid evaluation.
 -}
private pure native fork "frege.runtime.Concurrent.fork" :: (Bool -> a) -> Bool


{--
 * [usage] @a `par` b@

 * Fork evaluation of @a@ and return @b@.

 -}
par ?a ?b = if fork (defer a) then b else undefined
    where defer ?a b = if b then a else a

{--
 * [usage] @f $| a@
 *
 * Equivalent to @a `par` f a@
 -}
f $| ?a = a `par` f a

{--
 * [usage] @parOp f a b@

 * Equivalent to @a `par` b `par` f a b@
 -}
parOp f ?a ?b = a `par` b `par` f a b

{--
 * [usage] @a :| as@

 * Equivalent to @parOp (:) a as@
 -}
?a :| ?as = parOp (:) a as

{--
    [usage] @mapP f xs@

    Like 'map', but applies the function parallel to mapping the rest of the list.
    It does therefore not work with infinite lists.
-}

mapP f (x:xs) = parOp (:) (f x) (mapP f xs)
mapP f [] = []
